#include "../../internal-include/GPU/gpu_internal.hpp"
#include <PSX/System/IOPorts/interrupt_io.hpp>
#include <PSX/System/syscalls.h>

namespace JabyEngine {
    namespace GPU {
        uint8_t Display :: current_id = 1; //< Setup will call exchange and set it to 0

        namespace internal {
            static uint8_t vsync_count = 0;

            static InterruptVerifierResult interrupt_verifier();
            static void interrupt_handler(uint32_t);

            InterrupCallback       callback          = {
                .next              = nullptr,
                .handler_function  = reinterpret_cast<InterruptHandler>(interrupt_handler),
                .verifier_function = interrupt_verifier
            };

            static InterruptVerifierResult interrupt_verifier() {
                if(Interrupt::is_irq(Interrupt::VBlank)) {
                    return InterruptVerifierResult::ExecuteHandler;
                }
                
                else {
                    return InterruptVerifierResult::SkipHandler;
                }
            }

            static void interrupt_handler(uint32_t) {
                if(vsync_count != 0xFF) {
                    vsync_count++;
                }

                Interrupt::ack_irq(Interrupt::VBlank);
                __syscall_ReturnFromException();
            }

            void Display :: exchange_buffer_and_display() {
                GPU::internal::set_draw_area(0, (PublicDisplay::Height*PublicDisplay::current_id));
                PublicDisplay::current_id ^= 1;
                GPU_IO::GP1 = GPU_IO::Command::DisplayArea(0, (PublicDisplay::Height*PublicDisplay::current_id));
            }

            void wait_vsync(uint8_t syncs) {
                volatile uint8_t& vsync_count = reinterpret_cast<volatile uint8_t&>(internal::vsync_count);

                if(internal::vsync_count >= syncs) {
                    syncs = internal::vsync_count + 1;
                }

                while(vsync_count < syncs);
                vsync_count = 0;
            }
        }

        #ifndef USE_NO$PSX
            void Display :: set_offset(uint16_t x, uint16_t y) {
                x += 78;
                y += 43;

                GPU_IO::GP1 = GPU_IO::Command::HorizontalDisplayRange((x << 3), (x + Display::Width) << 3);
                GPU_IO::GP1 = GPU_IO::Command::VerticalDisplayRange(y,         y + Display::Height);
            }
        #else
            void Display :: set_offset(uint16_t x, uint16_t y) {
                GPU_IO::GP1 = GPU_IO::Command::HorizontalDisplayRange(x, (x + Display::Width*8));
                GPU_IO::GP1 = GPU_IO::Command::VerticalDisplayRange(y - (ScanlinesV/2), y + (ScanlinesV/2));
            }
        #endif //USE_NO$PSX

        uint8_t swap_buffers_vsync(uint8_t syncs) {
            // Wait finish draw
            internal::wait_vsync(syncs);
            internal::Display::exchange_buffer_and_display();
            return Display::current_id;
        }
    }
}